@@ -59,9 +59,12 @@ private Q_SLOTS:
     void testCapabilities_data();
     void testCapabilities();
     void testPointer();
+    void testPointerTransformation_data();
+    void testPointerTransformation();
     void testPointerButton_data();
     void testPointerButton();
     void testCursor();
+    void testCursorDamage();
     void testKeyboard();
     void testCast();
     void testDestroy();
@@ -294,8 +297,13 @@ void TestWaylandSeat::testPointer()
     SurfaceInterface *serverSurface = surfaceCreatedSpy.first().first().value<KWayland::Server::SurfaceInterface*>();
     QVERIFY(serverSurface);
 
+    QSignalSpy focusedPointerChangedSpy(m_seatInterface, &SeatInterface::focusedPointerChanged);
+    QVERIFY(focusedPointerChangedSpy.isValid());
+
     m_seatInterface->setPointerPos(QPoint(20, 18));
     m_seatInterface->setFocusedPointerSurface(serverSurface, QPoint(10, 15));
+    QCOMPARE(focusedPointerChangedSpy.count(), 1);
+    QVERIFY(!focusedPointerChangedSpy.first().first().value<PointerInterface*>());
     // no pointer yet
     QVERIFY(m_seatInterface->focusedPointerSurface());
     QVERIFY(!m_seatInterface->focusedPointer());
@@ -309,8 +317,12 @@ void TestWaylandSeat::testPointer()
     QVERIFY(pointerCreatedSpy.wait());
     QVERIFY(m_seatInterface->focusedPointer());
     QCOMPARE(pointerCreatedSpy.first().first().value<PointerInterface*>(), m_seatInterface->focusedPointer());
+    QCOMPARE(focusedPointerChangedSpy.count(), 2);
+    QCOMPARE(focusedPointerChangedSpy.last().first().value<PointerInterface*>(), m_seatInterface->focusedPointer());
 
     m_seatInterface->setFocusedPointerSurface(nullptr);
+    QCOMPARE(focusedPointerChangedSpy.count(), 3);
+    QVERIFY(!focusedPointerChangedSpy.last().first().value<PointerInterface*>());
     serverSurface->client()->flush();
     QTest::qWait(100);
 
@@ -340,6 +352,8 @@ void TestWaylandSeat::testPointer()
     QVERIFY(serverPointer);
     QCOMPARE(p->enteredSurface(), s);
     QCOMPARE(cp.enteredSurface(), s);
+    QCOMPARE(focusedPointerChangedSpy.count(), 4);
+    QCOMPARE(focusedPointerChangedSpy.last().first().value<PointerInterface*>(), serverPointer);
 
     // test motion
     m_seatInterface->setTimestamp(1);
@@ -409,6 +423,7 @@ void TestWaylandSeat::testPointer()
 
     // leave the surface
     m_seatInterface->setFocusedPointerSurface(nullptr);
+    QCOMPARE(focusedPointerChangedSpy.count(), 5);
     QVERIFY(leftSpy.wait());
     QCOMPARE(leftSpy.first().first().value<quint32>(), m_display->serial());
     QVERIFY(!p->enteredSurface());
@@ -416,6 +431,113 @@ void TestWaylandSeat::testPointer()
 
     // enter it again
     m_seatInterface->setFocusedPointerSurface(serverSurface, QPoint(0, 0));
+    QCOMPARE(focusedPointerChangedSpy.count(), 6);
+    QVERIFY(enteredSpy.wait());
+    QCOMPARE(p->enteredSurface(), s);
+    QCOMPARE(cp.enteredSurface(), s);
+
+    delete s;
+    wl_display_flush(m_connection->display());
+    QVERIFY(focusedPointerChangedSpy.wait());
+    QCOMPARE(focusedPointerChangedSpy.count(), 7);
+    QVERIFY(!m_seatInterface->focusedPointerSurface());
+}
+
+void TestWaylandSeat::testPointerTransformation_data()
+{
+    QTest::addColumn<QMatrix4x4>("enterTransformation");
+    // global position at 20/18
+    QTest::addColumn<QPointF>("expectedEnterPoint");
+    // global position at 10/16
+    QTest::addColumn<QPointF>("expectedMovePoint");
+
+    QMatrix4x4 tm;
+    tm.translate(-10, -15);
+    QTest::newRow("translation") << tm << QPointF(10, 3) << QPointF(0, 1);
+    QMatrix4x4 sm;
+    sm.scale(2, 2);
+    QTest::newRow("scale") << sm << QPointF(40, 36) << QPointF(20, 32);
+    QMatrix4x4 rotate;
+    rotate.rotate(90, 0, 0, 1);
+    QTest::newRow("rotate") << rotate << QPointF(-18, 20) << QPointF(-16, 10);
+}
+
+void TestWaylandSeat::testPointerTransformation()
+{
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+
+    QSignalSpy pointerSpy(m_seat, &Seat::hasPointerChanged);
+    QVERIFY(pointerSpy.isValid());
+    m_seatInterface->setHasPointer(true);
+    QVERIFY(pointerSpy.wait());
+
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, &CompositorInterface::surfaceCreated);
+    QVERIFY(surfaceCreatedSpy.isValid());
+    Surface *s = m_compositor->createSurface(m_compositor);
+    QVERIFY(surfaceCreatedSpy.wait());
+    SurfaceInterface *serverSurface = surfaceCreatedSpy.first().first().value<KWayland::Server::SurfaceInterface*>();
+    QVERIFY(serverSurface);
+
+    m_seatInterface->setPointerPos(QPoint(20, 18));
+    QFETCH(QMatrix4x4, enterTransformation);
+    m_seatInterface->setFocusedPointerSurface(serverSurface, enterTransformation);
+    QCOMPARE(m_seatInterface->focusedPointerSurfaceTransformation(), enterTransformation);
+    // no pointer yet
+    QVERIFY(m_seatInterface->focusedPointerSurface());
+    QVERIFY(!m_seatInterface->focusedPointer());
+
+    Pointer *p = m_seat->createPointer(m_seat);
+    const Pointer &cp = *p;
+    QVERIFY(p->isValid());
+    QSignalSpy pointerCreatedSpy(m_seatInterface, &SeatInterface::pointerCreated);
+    QVERIFY(pointerCreatedSpy.isValid());
+    // once the pointer is created it should be set as the focused pointer
+    QVERIFY(pointerCreatedSpy.wait());
+    QVERIFY(m_seatInterface->focusedPointer());
+    QCOMPARE(pointerCreatedSpy.first().first().value<PointerInterface*>(), m_seatInterface->focusedPointer());
+
+    m_seatInterface->setFocusedPointerSurface(nullptr);
+    serverSurface->client()->flush();
+    QTest::qWait(100);
+
+    QSignalSpy enteredSpy(p, &Pointer::entered);
+    QVERIFY(enteredSpy.isValid());
+
+    QSignalSpy leftSpy(p, &Pointer::left);
+    QVERIFY(leftSpy.isValid());
+
+    QSignalSpy motionSpy(p, &Pointer::motion);
+    QVERIFY(motionSpy.isValid());
+
+    QVERIFY(!p->enteredSurface());
+    QVERIFY(!cp.enteredSurface());
+    m_seatInterface->setFocusedPointerSurface(serverSurface, enterTransformation);
+    QCOMPARE(m_seatInterface->focusedPointerSurface(), serverSurface);
+    QVERIFY(enteredSpy.wait());
+    QCOMPARE(enteredSpy.first().first().value<quint32>(), m_display->serial());
+    QTEST(enteredSpy.first().last().toPointF(), "expectedEnterPoint");
+    PointerInterface *serverPointer = m_seatInterface->focusedPointer();
+    QVERIFY(serverPointer);
+    QCOMPARE(p->enteredSurface(), s);
+    QCOMPARE(cp.enteredSurface(), s);
+
+    // test motion
+    m_seatInterface->setTimestamp(1);
+    m_seatInterface->setPointerPos(QPoint(10, 16));
+    QVERIFY(motionSpy.wait());
+    QTEST(motionSpy.first().first().toPointF(), "expectedMovePoint");
+    QCOMPARE(motionSpy.first().last().value<quint32>(), quint32(1));
+
+    // leave the surface
+    m_seatInterface->setFocusedPointerSurface(nullptr);
+    QVERIFY(leftSpy.wait());
+    QCOMPARE(leftSpy.first().first().value<quint32>(), m_display->serial());
+    QVERIFY(!p->enteredSurface());
+    QVERIFY(!cp.enteredSurface());
+
+    // enter it again
+    m_seatInterface->setFocusedPointerSurface(serverSurface);
     QVERIFY(enteredSpy.wait());
     QCOMPARE(p->enteredSurface(), s);
     QCOMPARE(cp.enteredSurface(), s);
@@ -610,14 +732,80 @@ void TestWaylandSeat::testCursor()
     QCOMPARE(surfaceChangedSpy.count(), 1);
     QCOMPARE(cursor->surface()->buffer()->data(), img);
 
-    p->hideCursor();
-    QVERIFY(surfaceChangedSpy.wait());
+    // and add another image to the surface
+    QImage blue(QSize(10, 20), QImage::Format_ARGB32);
+    blue.fill(Qt::blue);
+    cursorSurface->attachBuffer(m_shm->createBuffer(blue));
+    cursorSurface->damage(QRect(0, 0, 10, 20));
+    cursorSurface->commit(Surface::CommitFlag::None);
+    QVERIFY(changedSpy.wait());
     QCOMPARE(changedSpy.count(), 4);
     QCOMPARE(cursorChangedSpy.count(), 5);
+    QCOMPARE(cursor->surface()->buffer()->data(), blue);
+
+    p->hideCursor();
+    QVERIFY(surfaceChangedSpy.wait());
+    QCOMPARE(changedSpy.count(), 5);
+    QCOMPARE(cursorChangedSpy.count(), 6);
     QCOMPARE(surfaceChangedSpy.count(), 2);
     QVERIFY(!cursor->surface());
 }
 
+void TestWaylandSeat::testCursorDamage()
+{
+    // this test verifies that damaging a cursor surface triggers a cursor changed on the server
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+
+    QSignalSpy pointerSpy(m_seat, &Seat::hasPointerChanged);
+    QVERIFY(pointerSpy.isValid());
+    m_seatInterface->setHasPointer(true);
+    QVERIFY(pointerSpy.wait());
+
+    // create pointer
+    QScopedPointer<Pointer> p(m_seat->createPointer());
+    QVERIFY(p->isValid());
+    QSignalSpy enteredSpy(p.data(), &Pointer::entered);
+    QVERIFY(enteredSpy.isValid());
+    // create surface
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, &CompositorInterface::surfaceCreated);
+    QVERIFY(surfaceCreatedSpy.isValid());
+    m_compositor->createSurface(m_compositor);
+    QVERIFY(surfaceCreatedSpy.wait());
+    SurfaceInterface *serverSurface = surfaceCreatedSpy.first().first().value<KWayland::Server::SurfaceInterface*>();
+    QVERIFY(serverSurface);
+
+    // send enter to the surface
+    m_seatInterface->setFocusedPointerSurface(serverSurface);
+    QVERIFY(enteredSpy.wait());
+
+    // create a signal spy for the cursor changed signal
+    auto pointer = m_seatInterface->focusedPointer();
+    QSignalSpy cursorChangedSpy(pointer, &PointerInterface::cursorChanged);
+    QVERIFY(cursorChangedSpy.isValid());
+
+    // now let's set the cursor
+    Surface *cursorSurface = m_compositor->createSurface(m_compositor);
+    QVERIFY(cursorSurface);
+    QImage red(QSize(10, 10), QImage::Format_ARGB32);
+    red.fill(Qt::red);
+    cursorSurface->attachBuffer(m_shm->createBuffer(red));
+    cursorSurface->damage(QRect(0, 0, 10, 10));
+    cursorSurface->commit(Surface::CommitFlag::None);
+    p->setCursor(cursorSurface, QPoint(0, 0));
+    QVERIFY(cursorChangedSpy.wait());
+    QCOMPARE(pointer->cursor()->surface()->buffer()->data(), red);
+
+    // and damage the surface
+    QImage blue(QSize(10, 10), QImage::Format_ARGB32);
+    blue.fill(Qt::blue);
+    cursorSurface->attachBuffer(m_shm->createBuffer(blue));
+    cursorSurface->damage(QRect(0, 0, 10, 10));
+    cursorSurface->commit(Surface::CommitFlag::None);
+    QVERIFY(cursorChangedSpy.wait());
+    QCOMPARE(pointer->cursor()->surface()->buffer()->data(), blue);
+}
+
 void TestWaylandSeat::testKeyboard()
 {
     using namespace KWayland::Client;
